unit sfmtrandom;

{***************************************************************************

  SIMD-oriented Fast Mersenne Twister 
   - SFMT-src-1.3.3.zip̈ڐA

  date 2008/11/02
  auth fT
  memo o
       32bit UINT݈̂ڐA

  date 2013/01/27
  auth fT
  memo SSE2ł̓ô߂ɓ\ύX

 ***************************************************************************}

interface

uses
  Winapi.windows,
  basernd;

const
  SFMT_MEXP = 19937;
  SFMT_N    = SFMT_MEXP div 128 + 1;
  SFMT_N32  = SFMT_N * 4;
  SFMT_N64  = SFMT_N * 2;
  SFMT_POS1 = 122;
  SFMT_SL1  = 18;
  SFMT_SL2  = 1;
  SFMT_SR1  = 11;
  SFMT_SR2  = 1;
  SFMT_MSK1 = $dfffffef;
  SFMT_MSK2 = $ddfecb7f;
  SFMT_MSK3 = $bffaffff;
  SFMT_MSK4 = $bffffff6;
  SFMT_PARITY1 = $00000001;
  SFMT_PARITY2 = $00000000;
  SFMT_PARITY3 = $00000000;
  SFMT_PARITY4 = $13c9e684;

type
  PW128_T = ^TW128_T;
  TW128_T = packed record
    u : array[0..3] of DWORD;
  end;

  PW128_T_ARY = ^TW128_T_ARY;
  TW128_T_ARY = array[0..SFMT_N-1] of TW128_T;

  PSFMTDWORDARRAY = ^TSFMTDWORDARRAY;
  TSFMTDWORDARRAY = array[0..SFMT_N32-1] of DWORD;

  TSFMTRnd = class(TRnd)
  private
    sfmt: TW128_T_ARY;
    psfmt32 : PSFMTDWORDARRAY;
    idx : integer;
    parity : array[0..3] of DWORD;
  protected
    procedure   Init; override;
    procedure   gen_rand_all; virtual;
    procedure   do_recursion(var r,a,b,c,d : TW128_T);
    procedure   lshift128(var out_v:TW128_T; in_v:TW128_T; shift:integer);
    procedure   rshift128(var out_v:TW128_T; in_v:TW128_T; shift:integer);
    function    func1(x:DWORD):DWORD; inline;
    function    func2(x:DWORD):DWORD; inline;
    procedure   period_certification;
    function    GetPSfmt(index:integer):PW128_T; virtual;
  public
    procedure   SetSeed(ASeed:DWORD); override;
    procedure   SetSeed(ASeeds:array of DWORD); override;
    function    Rand:DWORD; override;
  end;

implementation

function    TSFMTRnd.GetPSfmt(index:integer):PW128_T;
begin
  result:=@sfmt[index];
end;

procedure   TSFMTRnd.Init;
begin
  psfmt32:=@(GetPSfmt(0)^.u[0]);
  parity[0]:=SFMT_PARITY1;
  parity[1]:=SFMT_PARITY2;
  parity[2]:=SFMT_PARITY3;
  parity[3]:=SFMT_PARITY4;
end;

procedure   TSFMTRnd.SetSeed(ASeed:DWORD);
var
  i : DWORD;
begin
  psfmt32^[0]:=ASeed;
  for i := 1 to SFMT_N32 - 1 do
    psfmt32^[i]:=1812433253 * (psfmt32^[i-1] XOR (psfmt32^[i-1] SHR 30)) + i;
  idx:=SFMT_N32;
  period_certification;
end;

procedure   TSFMTRnd.SetSeed(ASeeds:array of DWORD);
var
  i,j,count,lag,mid,size,key_length,
  r : DWORD;
begin
  size:=SFMT_N*4;
  if size>=623 then
    lag:=11
  else if size>=68 then
    lag:=7
  else if size>=39 then
    lag:=5
  else
    lag:=3;
  mid:=(size-lag) div 2;

  FillChar(GetPSfmt(0)^.u[0],sizeof(sfmt),$8B);
  key_length:=High(ASeeds)+1;
  if (key_length+1)>SFMT_N32 then
    count:=key_length+1
  else
    count:=SFMT_N32;
  r:=func1(psfmt32^[0] XOR psfmt32^[mid] XOR psfmt32^[SFMT_N32-1]);
  psfmt32^[mid]:=psfmt32^[mid]+r;
  r:=r+key_length;
  psfmt32^[mid+lag]:=psfmt32^[mid+lag]+r;
  psfmt32^[0]:=r;

  Dec(count);
  i:=1; j:=0;
  while (j<count) and (j<key_length) do begin
    r:=func1(psfmt32^[i] XOR psfmt32^[(i+mid) mod SFMT_N32]
                         XOR psfmt32^[(i+SFMT_N32-1) mod SFMT_N32]);
    psfmt32^[(i+mid) mod SFMT_N32]:=psfmt32^[(i+mid) mod SFMT_N32]+r;
    r:=r+ASeeds[j]+i;
    psfmt32^[(i+mid+lag) mod SFMT_N32]:=psfmt32^[(i+mid+lag) mod SFMT_N32]+r;
    psfmt32^[i]:=r;
    i:=(i+1) mod SFMT_N32;
    Inc(j);
  end;

  while j<count do begin
    r:=func1(psfmt32^[i] XOR psfmt32^[(i+mid) mod SFMT_N32]
                         XOR psfmt32^[(i+SFMT_N32-1) mod SFMT_N32]);
    psfmt32^[(i+mid) mod SFMT_N32]:=psfmt32^[(i+mid) mod SFMT_N32]+r;
    r := r+i;
    psfmt32^[(i+mid+lag) mod SFMT_N32]:=psfmt32^[(i+mid+lag) mod SFMT_N32]+r;
    psfmt32^[i]:=r;
    i:=(i+1) mod SFMT_N32;
    Inc(j);
  end;

  for j:=0 to SFMT_N32-1 do begin
    r:=func2(psfmt32^[i] + psfmt32^[(i+mid) mod SFMT_N32]
                         + psfmt32^[(i+SFMT_N32-1) mod SFMT_N32]);
    psfmt32^[(i+mid) mod SFMT_N32]:=psfmt32^[(i+mid) mod SFMT_N32] XOR r;
    r:=r-i;
    psfmt32^[(i+mid+lag) mod SFMT_N32]:=psfmt32^[(i+mid+lag) mod SFMT_N32] XOR r;
    psfmt32^[i]:=r;
    i:=(i+1) mod SFMT_N32;
  end;

  idx:=SFMT_N32;
  period_certification;
end;

function    TSFMTRnd.func1(x:DWORD):DWORD;
begin
  result:= (x XOR (x SHR 27)) * 1664525;
end;

function    TSFMTRnd.func2(x:DWORD):DWORD;
begin
  result:= (x XOR (x SHR 27)) * 1566083941;
end;

procedure   TSFMTRnd.period_certification;
var
  i,j : integer;
  inner,work : DWORD;
begin
  inner:=0;
  for i := 0 to 3 do
    inner := inner XOR (psfmt32^[i] and parity[i]);
  i:=16;
  while i>0 do begin
    inner := inner XOR (inner SHR i);
    i:=i SHR 1;
  end;
  {* Check OK *}
  if inner=1 then exit;
  {* Check NG *}
  for i := 0 to 3 do begin
    work:=1;
    for j := 0 to 31 do begin
      if (work AND parity[i])<>0 then begin
        psfmt32^[i] := psfmt32^[i] XOR work;
        exit;
      end;
      work:= work SHL 1;
    end;
  end;
end;

procedure   TSFMTRnd.lshift128(var out_v:TW128_T; in_v:TW128_T; shift:integer);
var
  th, tl, oh, ol : int64;
begin
  th := (int64(in_v.u[3]) SHL 32) OR (int64(in_v.u[2]));
  tl := (int64(in_v.u[1]) SHL 32) OR (int64(in_v.u[0]));

  oh := th SHL (shift*8);
  ol := tl SHL (shift*8);
  oh := oh OR (tl SHR (64-shift*8));
  out_v.u[1] := integer(ol SHR 32);
  out_v.u[0] := integer(ol AND $FFFFFFFF);
  out_v.u[3] := integer(oh SHR 32);
  out_v.u[2] := integer(oh AND $FFFFFFFF);
end;

procedure   TSFMTRnd.rshift128(var out_v:TW128_T; in_v:TW128_T; shift:integer);
var
  th, tl, oh, ol : int64;
begin
  th := (int64(in_v.u[3]) SHL 32) OR (int64(in_v.u[2]));
  tl := (int64(in_v.u[1]) SHL 32) OR (int64(in_v.u[0]));

  oh := th SHR (shift*8);
  ol := tl SHR (shift*8);
  ol := ol OR (th SHL (64-shift*8));
  out_v.u[1] := integer(ol SHR 32);
  out_v.u[0] := integer(ol AND $FFFFFFFF);
  out_v.u[3] := integer(oh SHR 32);
  out_v.u[2] := integer(oh AND $FFFFFFFF);
end;

procedure   TSFMTRnd.do_recursion(var r,a,b,c,d : TW128_T);
var
  x,y : TW128_T;
begin
  lshift128(x,a,SFMT_SL2);
  rshift128(y,c,SFMT_SR2);
  r.u[0]:=a.u[0] XOR x.u[0] XOR ((b.u[0] SHR SFMT_SR1) AND SFMT_MSK1) XOR y.u[0] XOR (d.u[0] SHL SFMT_SL1);
  r.u[1]:=a.u[1] XOR x.u[1] XOR ((b.u[1] SHR SFMT_SR1) AND SFMT_MSK2) XOR y.u[1] XOR (d.u[1] SHL SFMT_SL1);
  r.u[2]:=a.u[2] XOR x.u[2] XOR ((b.u[2] SHR SFMT_SR1) AND SFMT_MSK3) XOR y.u[2] XOR (d.u[2] SHL SFMT_SL1);
  r.u[3]:=a.u[3] XOR x.u[3] XOR ((b.u[3] SHR SFMT_SR1) AND SFMT_MSK4) XOR y.u[3] XOR (d.u[3] SHL SFMT_SL1);
end;

procedure   TSFMTRnd.gen_rand_all;
var
  i : integer;
  r1,r2 : PW128_T;
begin
  r1:=@sfmt[SFMT_N-2];
  r2:=@sfmt[SFMT_N-1];
  for i := 0 to SFMT_N-SFMT_POS1-1 do begin
    do_recursion(sfmt[i],sfmt[i],sfmt[i+SFMT_POS1],r1^,r2^);
    r1:=r2;
    r2:=@sfmt[i];
  end;
  for i := SFMT_N-SFMT_POS1 to SFMT_N - 1 do  begin
    do_recursion(sfmt[i],sfmt[i],sfmt[i+SFMT_POS1-SFMT_N],r1^,r2^);
    r1:=r2;
    r2:=@sfmt[i];
  end;
end;

function    TSFMTRnd.Rand:DWORD;
begin
  if idx>=SFMT_N32 then begin
    gen_rand_all;
    idx:=0;
  end;
  result:=psfmt32^[idx];
  Inc(idx);
end;

end.
