﻿unit sfmtsse2random;

{***************************************************************************

  SSE2版SIMD-oriented Fast Mersenne Twister 乱数生成器
   - SFMT-src-1.3.3.zipからの移植

  date 2013/01/27
  auth 綾川鞠乃介
  memo 初出

  date 2013/01/28
  auth 綾川鞠乃介
  memo Maskの初期化をInitで行うように変更
       gen_rand_allのpush ?ax/pop ?ax除去
       64bit版のミスを修正(たまたま実害は無かったけど...)

  date 2013/01/29
  auth 綾川鞠乃介
  memo SSE2版がまともに動いていなかったのを修正
       今度は検証コード "も" 確認....

  date 2013/01/31
  auth 綾川鞠乃介
  memo gen_rand_allに $CODEALIGN 16 追加

  date 2015/10/24
  auth 綾川鞠乃介
  memo Dlephi 10用にxmmレジスタの保護範囲を広げる(64bit用)

 ***************************************************************************}

interface

uses
  Winapi.windows,
  basernd,sfmtrandom;

type
  TSFMTSSE2Rnd = class(TSfmtRnd)
  private
    pDATA : pointer;
    psfmt: PW128_T_ARY;
    pMask : PW128_T;
  protected
    procedure   Init; override;
    function    GetPSfmt(index:integer):PW128_T; override;
    procedure   gen_rand_all; override;
  public
    destructor  Destroy; override;
  end;

implementation

const
  AdjustMemAlignValue = 16;

destructor  TSFMTSSE2Rnd.Destroy;
begin
  IF pData<>nil then FreeMem(pData);
  inherited Destroy;
end;

procedure   TSFMTSSE2Rnd.Init;
var
  i : NativeUInt;
begin
  pData:=nil;
  // Adjust 16byte align for SSE2
  GetMem(pData,sizeof(TW128_T_ARY)+sizeof(TW128_T)+AdjustMemAlignValue);
  i:=16-NativeInt(NativeUInt(pData) AND $0F);
  if i=16 then i:=0;
  psfmt:=PW128_T_ARY(NativeUInt(pData)+i);
  pMask:=PW128_T(NativeUInt(psfmt)+sizeof(TW128_T_ARY));
  pMask.u[0]:=SFMT_MSK1;
  pMask.u[1]:=SFMT_MSK2;
  pMask.u[2]:=SFMT_MSK3;
  pMask.u[3]:=SFMT_MSK4;
  //
  inherited Init;
end;

function    TSFMTSSE2Rnd.GetPSfmt(index:integer):PW128_T;
begin
  result:=@(psfmt^[index]);
end;

// VC11 Update1 で asmソースをはき出してそれを手直ししてDelphiに移植
{$CODEALIGN 16}
procedure   TSFMTSSE2Rnd.gen_rand_all;
const
  PRMSZ = sizeof(TW128_T);  // assembler中に直接sizeof(TW128_T)を記述すると不正な値に...
asm
{$IFDEF WIN64}
  .SAVENV xmm4
  .NOFRAME
  // self = rcx
  mov         rdx, [rcx+pMask]
  mov         rax, [rcx+psfmt]
  movdqa      xmm4, [rdx]                       // mask = _mm_set_epi32(MSK4, MSK3, MSK2, MSK1);
  movdqa      xmm2, [rax+(SFMT_N-2)*PRMSZ]      // r1 = _mm_load_si128(&sfmt[N - 2].si);
  movdqa      xmm3, [rax+(SFMT_N-1)*PRMSZ]      // r2 = _mm_load_si128(&sfmt[N - 1].si);
  mov         rdx, rax
  add         rdx, (SFMT_N-SFMT_POS1)*PRMSZ
@@loop1:
  movdqa      xmm1, dqword ptr [rax]            //   x = _mm_load_si128(a);
  psrldq      xmm2, SFMT_SR2                    //   z = _mm_srli_si128(c, SR2);
  movdqa      xmm0, xmm3
  pxor        xmm2, xmm1                        //   z = _mm_xor_si128(z, x);
  pslld       xmm0, SFMT_SL1                    //   v = _mm_slli_epi32(d, SL1);
  pslldq      xmm1, SFMT_SL2                    //   x = _mm_slli_si128(x, SL2);
  pxor        xmm2, xmm0                        //   z = _mm_xor_si128(z, v);
  movdqa      xmm0, dqword ptr [rax+SFMT_POS1*PRMSZ]
  psrld       xmm0, SFMT_SR1                    //   y = _mm_srli_epi32(*b, SR1);
  pxor        xmm1, xmm2                        //   z = _mm_xor_si128(z, x);
  pand        xmm0, xmm4                        //   y = _mm_and_si128(y, mask);
  pxor        xmm1, xmm0                        //   z = _mm_xor_si128(z, y);
  movdqa      dqword ptr [rax], xmm1            // _mm_store_si128(&sfmt[i].si, r);
  add         rax, PRMSZ
  movdqa      xmm2, xmm3
  movdqa      xmm3, xmm1
  cmp         rax, rdx
  jne         @@loop1
  add         rdx, SFMT_POS1*PRMSZ
@@loop2:
  movdqa      xmm0, dqword ptr [rax]
  psrldq      xmm2, SFMT_SR2
  pslld       xmm1, SFMT_SL1
  pxor        xmm2, xmm0
  pslldq      xmm0, SFMT_SL2
  pxor        xmm1, xmm2
  pxor        xmm1, xmm0
  movdqa      xmm0, dqword ptr [rax-(SFMT_N-SFMT_POS1)*PRMSZ]
  psrld       xmm0, SFMT_SR1
  add         rax, PRMSZ
  pand        xmm0, xmm4
  pxor        xmm1, xmm0
  movdqa      dqword ptr [rax-PRMSZ], xmm1
  movdqa      xmm2, xmm3
  movdqa      xmm3, xmm1
  cmp         rax, rdx
  jne         @@loop2
{$ELSE}
  // self = eax
  mov         ecx, [eax+pMask]
  mov         edx, [eax+psfmt]
  movdqa      xmm4, [ecx]                       // mask = _mm_set_epi32(MSK4, MSK3, MSK2, MSK1);
  movdqa      xmm2, [edx+(SFMT_N-2)*PRMSZ]      // r1 = _mm_load_si128(&sfmt[N - 2].si);
  movdqa      xmm3, [edx+(SFMT_N-1)*PRMSZ]      // r2 = _mm_load_si128(&sfmt[N - 1].si);
  mov         ecx, edx
  add         ecx, (SFMT_N-SFMT_POS1)*PRMSZ
@@loop1:
  movdqa      xmm1, dqword ptr [edx]            //   x = _mm_load_si128(a);
  psrldq      xmm2, SFMT_SR2                    //   z = _mm_srli_si128(c, SR2);
  movdqa      xmm0, xmm3
  pxor        xmm2, xmm1                        //   z = _mm_xor_si128(z, x);
  pslld       xmm0, SFMT_SL1                    //   v = _mm_slli_epi32(d, SL1);
  pslldq      xmm1, SFMT_SL2                    //   x = _mm_slli_si128(x, SL2);
  pxor        xmm2, xmm0                        //   z = _mm_xor_si128(z, v);
  movdqa      xmm0, dqword ptr [edx+SFMT_POS1*PRMSZ]
  psrld       xmm0, SFMT_SR1                    //   y = _mm_srli_epi32(*b, SR1);
  pxor        xmm1, xmm2                        //   z = _mm_xor_si128(z, x);
  pand        xmm0, xmm4                        //   y = _mm_and_si128(y, mask);
  pxor        xmm1, xmm0                        //   z = _mm_xor_si128(z, y);
  movdqa      dqword ptr [edx], xmm1            // _mm_store_si128(&sfmt[i].si, r);
  add         edx, PRMSZ
  movdqa      xmm2, xmm3
  movdqa      xmm3, xmm1
  cmp         edx, ecx
  jne         @@loop1
  add         ecx, SFMT_POS1*PRMSZ
@@loop2:
  movdqa      xmm0, dqword ptr [edx]
  psrldq      xmm2, SFMT_SR2
  pslld       xmm1, SFMT_SL1
  pxor        xmm2, xmm0
  pslldq      xmm0, SFMT_SL2
  pxor        xmm1, xmm2
  pxor        xmm1, xmm0
  movdqa      xmm0, dqword ptr [edx-(SFMT_N-SFMT_POS1)*PRMSZ]
  psrld       xmm0, SFMT_SR1
  add         edx, PRMSZ
  pand        xmm0, xmm4
  pxor        xmm1, xmm0
  movdqa      dqword ptr [edx-PRMSZ], xmm1
  movdqa      xmm2, xmm3
  movdqa      xmm3, xmm1
  cmp         edx,  ecx
  jne         @@loop2
{$ENDIF}
end;


end.
